import pandas as pd
from tqdm import tqdm
import anndata as ad
import archs4py as a4
import os
import traceback
from .rnaseq_analysis import RNASeqAnalysis
from .rag_embedding import Rag_embedding
from .transcriptome_embedding import Transcriptome_embedding
from .enrichment import Transcriptome_enrichment

os.environ["TOKENIZERS_PARALLELISM"] = "false"
import warnings
import logging
import sys
warnings.filterwarnings('ignore')
 
logging.basicConfig(filename='log.txt', level=logging.INFO, format='%(asctime)s %(levelname)s %(message)s')

class Query_DB:
    logger = logging.getLogger(__name__)
    
    def __init__(self, semantic_vector_store, transcriptomic_vector_store, h5file=None):
        self.logger = logging.getLogger(__name__)
        H = logging.StreamHandler(sys.stdout)
        H.setLevel(logging.INFO)
        H.setFormatter(
            logging.Formatter(
                fmt="[%(asctime)s] %(levelname)s: %(message)s",
                datefmt="%d/%m/%Y ( %H:%M:%S )"
            ))
        self.logger.addHandler(H)

        self.logger.info("Loading transcriptomic vector store...")
        trans_obj = ad.read_h5ad(transcriptomic_vector_store, backed="r")
        self.logger.info("Transcriptomic vector store loaded.")
        
        self.logger.info("Loading semantic vector store...")
        sem_obj = ad.read_h5ad(semantic_vector_store, backed="r")
        self.logger.info("Semantic vector store loaded.")
        
        self.transcriptome_embedding = Transcriptome_embedding(trans_obj.obs, trans_obj.obsm["embedding"])
        self.logger.info("Transcriptome embedding initialized.")
        self.rag_embedding = Rag_embedding(sem_obj.obs, sem_obj.X)
        self.logger.info("RAG embedding initialized.")
        self.transcriptome_enrichment = Transcriptome_enrichment(trans_obj)        
        self.logger.info("Transcriptome object initialized.")
        
        if h5file is not None:
            self.logger.info("Loading ARCHS4 database object...")
            self.RNASeqAnalysis = RNASeqAnalysis(h5file)
            self.h5file = h5file    
            self.logger.info("ARCHS4 database object loaded.")
            
            self.logger.info("Creating series to sample mapping...")
            self.metafile = self.process_metadata(h5file)
            self.logger.info("DONE.")
        else:
            self.logger.info("ARCHS4 database object not loaded.")
            self.RNASeqAnalysis = None
        
        self.logger.info("Query_DB object initialized.")

    def process_metadata(self, h5_path):
        """
        Process metadata from an HDF5 file.

        Args:
            h5_path (str): The path to the HDF5 file.

        Returns:
            pandas.DataFrame: A DataFrame containing processed metadata with columns 'series_id' and 'samples'.
        """
        x = a4.meta.meta(h5_path, ".*", meta_fields=["series_id"])
        x["samples"] = x.index
        x = x[['series_id', 'samples']].drop_duplicates()
        return x

    def retrieve_sample_metadata_from_h5(self, h5_path, sample_list):
        """
        Retrieves sample metadata from an HDF5 file.

        Args:
            h5_path (str): The path to the HDF5 file.
            sample_list (list): A list of sample names.

        Returns:
            dict: A dictionary containing the sample metadata.

        """
        sample_meta = a4.meta.samples(h5_path, sample_list)
        return sample_meta

    def get_top_samples(self, df, n=1000):
        """
        Returns a subset of samples from the given DataFrame based on their ranking.

        Parameters:
        - df (pandas.DataFrame): The DataFrame containing the samples.
        - n (int): The number of top samples to retrieve. Default is 1000.

        Returns:
        - pandas.DataFrame: A subset of the original DataFrame containing the top samples.
        """
        df_meta = self.transcriptome_embedding.embeddings_index.copy()
        samps = df.iloc[:, 1].sort_values(ascending=False).head(n).index.to_list()
        series_of_interest = df_meta[df_meta.index.isin(samps)]
        return series_of_interest

    def transcriptome_search(self, geneset, nsamples=1000):
        """
        Perform a transcriptome search using a given geneset.

        Args:
            geneset (list): A list of genes to search for in the transcriptome.
            nsamples (int, optional): The number of top samples to retrieve. Defaults to 1000.

        Returns:
            tuple: A tuple containing two pandas DataFrames. The first DataFrame contains the top samples of interest,
                   and the second DataFrame contains the relevant series (studies) associated with the top samples.
        """
        my_list = self.transcriptome_enrichment.memmap_adata.obs.index
        user_batch_size = 500
        output_df1_list = []
        self.logger.info("Starting transcriptome search...")
        for i in tqdm(range(0, len(my_list), user_batch_size)):
            try:
                samps = my_list[i:i + user_batch_size]
                df1 = self.transcriptome_enrichment.run_decouplr_on_memmaped_adata_with_samples(geneset, samps)
                # Store the output from each iteration
                output_df1_list.append(df1)
            except Exception as e:
                self.logger.error(f"Failure in batch {i}: {str(e)}")
                self.logger.error(traceback.format_exc())
        final_df1 = pd.concat(output_df1_list)
        series_of_interest = self.get_top_samples(final_df1, n=nsamples)
        relevant_series = series_of_interest["series_id"]  # extract only studies (series)
        series_of_interest = series_of_interest.rename(columns={"series_id": "gse_id"})
        return series_of_interest, relevant_series
        
    
    def semantic_search(self, test_text, k=50):
        """
        Perform semantic search using the RAG embedding model.

        Parameters:
        - test_text (str): The text to search for.
        - k (int): The number of results to retrieve (default is 50).

        Returns:
        - df_res (DataFrame): The search results as a DataFrame.
        - df_res["gse_id"] (Series): The series containing the "gse_id" column from the search results.
        """
        self.logger.info("Starting semantic search...")
        df_res = self.rag_embedding.query_rag(test_text, k=k)
        return df_res, df_res["gse_id"]
 
    def get_transcriptome_series_of_relevance_from_series(self, series_of_interest, k=5):
        """
        Retrieves a series of transcriptome data that are relevant to the given series of interest.

        Parameters:
        - series_of_interest (list): A list of series of interest.
        - k (int): The number of closest transcriptional studies to retrieve for each series.

        Returns:
        - additional_series (DataFrame): A DataFrame containing the additional series of transcriptome data.
        """
        self.logger.info("Performing transcriptome expansion...")
        series_list = []
        target_series = series_of_interest
        for i in tqdm(target_series):
            res_temp = self.transcriptome_embedding.get_closest_transcriptional_studies(i, k=k)
            series_list.append(res_temp)
        additional_series = pd.concat(series_list)
        additional_series["gse_id_text"] = [self.rag_embedding.get_text_linked_to_gse(i) for i in additional_series["gse_id"]]
        return additional_series
    
    def get_semantic_series_of_relevance_from_series(self, series_of_interest, k=5):
        """
        Retrieves additional series of relevance based on the input series of interest.

        Parameters:
        - series_of_interest (list): A list of series of interest.
        - k (int): The number of closest semantic studies to retrieve for each series.

        Returns:
        - additional_series (pandas.DataFrame): A DataFrame containing the additional series of relevance.

        Usage:
        Call this method to retrieve additional series of relevance based on the input series of interest.
        The method uses the rag_embedding object to get the closest semantic studies for each series.
        The number of closest semantic studies to retrieve can be specified using the k parameter.
        """
        self.logger.info("Performing semantic expansion...")
        additional_series = pd.concat([self.rag_embedding.get_closest_semantic_studies(i, k=k) for i in series_of_interest])
        return additional_series
    

    def transcriptome_search_with_semantic_expansion(self, geneset_query, search=1000, expand=5):
        """
        Perform a transcriptome search with semantic expansion.

        Args:
            geneset_query (str): The query for the geneset.
            search (int, optional): The number of samples to search. Defaults to 1000.
            expand (int, optional): The number of additional series to expand the search. Defaults to 5.

        Returns:
            tuple or None: A tuple containing the additional series and the series dataframe if expand is not 0,
            otherwise None and the series dataframe.
        """
        self.logger.info("Starting transcriptome search with semantic expansion...")
        self.logger.info("search: " + str(search)),
        self.logger.info("expand: " + str(expand))
        if expand == 0:
            series_df, series_of_interest = self.transcriptome_search(geneset_query, search)
            return None, series_df
        else:
            series_df, series_of_interest = self.transcriptome_search(geneset=geneset_query, nsamples=search)
            additional_series = self.get_semantic_series_of_relevance_from_series(series_of_interest, expand)
            return additional_series, series_df

    def transcriptome_search_with_transcriptome_expansion(self, geneset_query, search=1000, expand=10):
        """
        Perform a transcriptome search with transcriptome expansion.

        Args:
            geneset_query (str): The geneset query to search for.
            search (int): The number of search results to retrieve. Default is 1000.
            expand (int): The number of additional series to expand the search. Default is 10.

        Returns:
            tuple or None: A tuple containing the additional series and the series dataframe if expand is not 0,
            otherwise None and the series dataframe.

        """
        self.logger.info("Starting transcriptome search with transcriptome expansion...")
        self.logger.info("search: " + str(search))
        self.logger.info("expand: " + str(expand))  
        if expand == 0:
            series_df, series_of_interest = self.transcriptome_search(geneset_query, search)
            return None, series_df
        else:
            series_df, series_of_interest = self.transcriptome_search(geneset_query, search)
            additional_series = self.get_transcriptome_series_of_relevance_from_series(series_of_interest, expand)
            return additional_series, series_df

    def semantic_search_with_semantic_expansion(self, text_query, search=50, expand=5):
        """
        Perform semantic search with semantic expansion.

        Args:
            text_query (str): The text query to search for.
            search (int): The number of search results to retrieve. Default is 50.
            expand (int): The number of additional series to expand the search results. Default is 5.

        Returns:
            tuple or None: A tuple containing the additional series and the search results dataframe.
                           If `expand` is 0, returns None as the additional series.
        """
        self.logger.info("Starting semantic search with semantic expansion...")
        self.logger.info("search: " + str(search))  
        self.logger.info("expand: " + str(expand))
        if expand == 0:
            series_df, series_of_interest = self.semantic_search(text_query, search)
            return None, series_df
        else:
            series_df, series_of_interest = self.semantic_search(text_query, search)
            additional_series = self.get_semantic_series_of_relevance_from_series(series_of_interest, expand)
            return additional_series, series_df

    def semantic_search_with_transcriptome_expansion(self, text_query, search=50, expand=10):
        """
        Performs semantic search with transcriptome expansion.

        Args:
            text_query (str): The text query to search for.
            search (int): The number of search results to retrieve.
            expand (int): The number of additional transcriptome series to expand the search with.

        Returns:
            tuple or None: A tuple containing the additional transcriptome series and the search results dataframe.
                           If `expand` is 0, returns `None` as the first element of the tuple.
        """
        self.logger.info("Starting semantic search with transcriptome expansion...")
        self.logger.info("search: " + str(search))
        self.logger.info("expand: " + str(expand))
        if expand == 0:
            series_df, series_of_interest = self.semantic_search(text_query, search)
            return None, series_df
        else:
            series_df, series_of_interest = self.semantic_search(text_query, search)
            additional_series = self.get_transcriptome_series_of_relevance_from_series(series_of_interest, expand)
            return additional_series, series_df

    def search(self, geneset, text_query, search="semantic", expand="transcriptome", perform_enrichment=False, n_seed=None, n_expansion=None):
        """
        Perform a search using the specified parameters.

        Args:
            geneset (list): A list of genes to search for.
            text_query (str): The text query to search for.
            search (str, optional): The type of search to perform. Defaults to "semantic".
            expand (str, optional): The type of expansion to perform. Defaults to "transcriptome".
            perform_enrichment (bool, optional): Whether to perform enrichment analysis. Defaults to False.
            n_seed (int, optional): The number of seed studies to include. Defaults to None.
            n_expansion (int, optional): The number of expansion studies to include. Defaults to None.

        Raises:
            ValueError: If the gene set is empty or has less than 5 genes.
            ValueError: If either n_seed or n_expansion is provided without the other.

        Returns:
            Results: An object containing the search results.
        """
        if geneset is None and text_query is None:
            raise ValueError("At least one of geneset or text_query should not be None.")
        
        if (n_seed is None and n_expansion is not None) or (n_seed is not None and n_expansion is None):
            raise ValueError("Both n_seed and n_expansion must be either None or filled.")
        
        if n_seed == 0:
            raise ValueError("n_seed should not be zero.")

        if geneset is not None:
            if len(geneset) < 5:
                self.logger.warning("Gene set should have at least 5 genes.")
        
            if (self.transcriptome_enrichment.memmap_adata.var["gene"].isin(geneset)).sum() == 0:
                raise ValueError("No genes from the gene set found in the transcriptome.")
            
            if (self.transcriptome_enrichment.memmap_adata.var["gene"].isin(geneset)).sum() < len(geneset):
                self.logger.warning("Some genes from the gene set were not found in the transcriptome.")
                missing_genes = set(geneset) - set(self.transcriptome_enrichment.memmap_adata.var["gene"])
                missing_genes_percentage = (len(missing_genes) / len(geneset)) * 100
                self.logger.warning(f"Missing genes from geneset in transcriptome: {missing_genes} ({missing_genes_percentage:.2f}% of geneset)")

        results_object = Results(None, None, None) # new results object

        if text_query is None:
            search = "transcriptome"
            self.logger.info("Search type set to transcriptome by default.")
            if geneset is None or len(geneset) == 0:
                raise ValueError("Gene set should not be empty.")

        if geneset is None:
            search = "semantic"
            self.logger.info("Search type set to semantic by default.")
            if text_query is None or len(text_query) == 0:
                raise ValueError("Text query should not be empty.")
            if perform_enrichment:
                self.logger.error("Enrichment not performed. No gene set provided.")
                raise ValueError("Please specify a gene set.")

        if search == "semantic":
            if expand == "transcriptome":
                if n_seed is None and n_expansion is None:
                    additional_series, seed_series = self.semantic_search_with_transcriptome_expansion(text_query)
                else:
                    additional_series, seed_series = self.semantic_search_with_transcriptome_expansion(text_query, search = n_seed, expand = n_expansion)
                results_object.seed_studies = seed_series
                results_object.expansion_studies = additional_series
            elif expand == "semantic":
                if n_seed is None and n_expansion is None:
                    additional_series, seed_series = self.semantic_search_with_semantic_expansion(text_query)
                else:
                    additional_series, seed_series = self.semantic_search_with_semantic_expansion(text_query, search = n_seed, expand = n_expansion)
                results_object.seed_studies = seed_series
                results_object.expansion_studies = additional_series
        if search == "transcriptome":
            if expand == "transcriptome":
                if n_seed is None and n_expansion is None:
                    additional_series, seed_series = self.transcriptome_search_with_transcriptome_expansion(geneset)
                else:
                    additional_series, seed_series = self.transcriptome_search_with_transcriptome_expansion(geneset, search = n_seed, expand = n_expansion)
                results_object.seed_studies = seed_series
                results_object.expansion_studies = additional_series
            elif expand == "semantic":
                if n_seed is None and n_expansion is None:
                    additional_series, seed_series = self.transcriptome_search_with_semantic_expansion(geneset)
                else :
                    additional_series, seed_series = self.transcriptome_search_with_semantic_expansion(geneset, search = n_seed, expand = n_expansion)
                results_object.seed_studies = seed_series
                results_object.expansion_studies = additional_series

        self.logger.info("Completed search step.")


        if perform_enrichment and self.RNASeqAnalysis is not None:
            if results_object.expansion_studies is not None:
                self.logger.info("Performing enrichment on expansion studies.")
                samps = self.metafile[self.metafile["series_id"].isin(results_object.expansion_studies["gse_id"])]
                enrichment_df = self.RNASeqAnalysis.perform_enrichment_on_samples_batched(samps.index, geneset)
                res_df = pd.concat([enrichment_df, samps], axis = 1)
                results_object.samples = res_df
                self.logger.info("Completed enrichment on studies in expansion step.")
            else:
                samps = self.metafile[self.metafile["series_id"].isin(results_object.seed_studies["gse_id"])]
                self.logger.info("No expansion studies found. Performing enrichment on seed studies.")
                enrichment_df = self.RNASeqAnalysis.perform_enrichment_on_samples_batched(samps.index, geneset)
                res_df = pd.concat([enrichment_df, samps], axis = 1)
                results_object.samples = res_df
                self.logger.info("Completed enrichment on seed studies.")
        elif perform_enrichment is False and self.RNASeqAnalysis is not None:
            if results_object.expansion_studies is not None:
                samps = self.metafile[self.metafile["series_id"].isin(results_object.expansion_studies["gse_id"])]
                res_df = self.retrieve_sample_metadata_from_h5(self.h5file, samps["samples"])
                results_object.samples = res_df
                self.logger.info("Retrieving samples from expansion step. Enrichment not performed.")
            else:
                samps = self.metafile[self.metafile["series_id"].isin(results_object.seed_studies["gse_id"])]
                res_df = self.retrieve_sample_metadata_from_h5(self.h5file, samps["samples"])
                results_object.samples = res_df
                self.logger.info("Retrieving samples from search step. Enrichment not performed.")
        else:
            self.logger.info("Enrichment not performed.")
            results_object.samples = None
            
        #additional_series = additional_series.drop(["similarity_score"], axis=1).reset_index(drop=True).drop_duplicates() 
        #seed_series = seed_series.drop_duplicates()
        self.logger.info("query result retrieved.")
        return results_object

class Results:
    """
    Represents the results of a bioRAG analysis.

    Attributes:
        seed_studies (DataFrame): The seed studies used in the analysis.
        expansion_studies (DataFrame): The expansion studies used in the analysis.
        samples (DataFrame): The samples used in the analysis.
    """

    def __init__(self, seed_df, expansion_df, samples_df):
        self.seed_studies = seed_df
        self.expansion_studies = expansion_df
        self.samples = samples_df
    



        




import argparse
from .biorag import Query_DB
import os
import datetime
import sys
import logging

def main():
    parser = argparse.ArgumentParser(description='BioRAG Command Line Interface')
    parser.add_argument('--gene_list', type=str, nargs='?', default=None, help='Path to the gene list text file or comma-separated gene list (e.g. "IRF1,IRF2,IRF3")')
    parser.add_argument('--text_query', type=str, nargs='?', default=None, help='Path to a text file or a string, e.g. "studies with cardiac myocytes"')
    parser.add_argument('--archs4_file', type=str, help='Path to the ARCHS4 hdf5 file')
    parser.add_argument('--transcriptome_db', type=str, help='Path to the transcriptome database')
    parser.add_argument('--semantic_db', type=str, help='Path to the vector database')
    parser.add_argument('--search', type=str, help='Search strategy', default='semantic')
    parser.add_argument('--expand', type=str, help='Expansion strategy', default='transcriptome')
    parser.add_argument('--enrichment', action='store_true', default=False, help='Enrichment parameter, if True, then performs ssGSEA')  # Added the --enrichment argument with default value False
    parser.add_argument('--usage', action='store_true', default=False, help='Print usage information')  # Added the --usage argument to print the usage information

    # Add any other command line arguments you need here

    parser.add_argument('--version', action='version', version='%(prog)s 0.1.5')  # Added the --version argument to print the version

    if '--usage' in sys.argv:
        print("=====================================================")
        print("BioRAG Command Line Interface")
        print("Version 0.1.7")
        print("Maintainer: WL Chin (melvin.chin@telethonkids.org.au)")
        print("=====================================================")
        print("")
        print("This is the usage information for the BioRAG Command Line Interface.")
        print("Please refer to the documentation on https://github.com/wlchin/bioRAG/ for detailed instructions on how to use this tool.")
        print("")
        print("Usage on command line:")
        print("----------------------")
        print("biorag --gene_list IRF1,IRF2,IRF3 --text_query 'studies with cardiac myocytes' --archs4_file ARCHS4.h5 --transcriptome_db transcriptome.h5 --semantic_db vector.h5")
        print("")
        print("Usage using files:")
        print("------------------")
        print("biorag --gene_list gene_list.txt --text_query text_query.txt --archs4_file ARCHS4.h5 --transcriptome_db transcriptome.h5 --semantic_db vector.h5")
        print("")
        print("Notes:")
        print("------")
        print("1. The gene names should be in uppercase")
        print("2. When using files, the gene_list.txt should contain a list of gene names (in UPPERCASE), one gene per line.")
        print("3. The file containing the text query should have the query string, which may contain multiple sentences, in plain text format")
        print("")    
        sys.exit(0)

    args = parser.parse_args()

    # Check if gene_list is provided
    if args.gene_list is not None:
        # Check if gene_list is a file or a comma-separated string
        if ',' in args.gene_list:
            # Split the comma-separated string into a list
            gene_list = args.gene_list.split(',')
        else:
            # Read the gene list from the file line by line and clean each line
            with open(args.gene_list, 'r') as gene_list_file:
                gene_list = [line.strip() for line in gene_list_file]
    else:
        gene_list = None

    # Check if text_query is provided
    if args.text_query is not None:
        # Check if text_query is a file or a string
        if args.text_query.endswith('.txt'):
            # Read the query from the file
            with open(args.text_query, 'r') as query_file:
                query = query_file.read()
        else:
            # Use the query as is
            query = args.text_query
    else:
        query = None

    new_query_db = Query_DB(args.semantic_db, args.transcriptome_db, args.archs4_file)
    result = new_query_db.search(geneset=gene_list, text_query=query, search=args.search, \
                                 expand=args.expand, perform_enrichment=args.enrichment)

    # Create the output folder if it doesn't exist
    output_folder = os.path.join('results', datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
        logging.info(f"Results folder created: {output_folder}")

    # Save the result to CSV files in the output folder
    result.seed_studies.to_csv(os.path.join(output_folder, "relevant_seed_studies.csv"))
    result.expansion_studies.to_csv(os.path.join(output_folder, "relevant_expansion_studies.csv"))
    result.samples.to_csv(os.path.join(output_folder, "relevant_samples.csv"))
    logging.info(f"Results saved to {output_folder}")

if __name__ == '__main__':
    main()



import pandas as pd
from sklearn.metrics.pairwise import cosine_similarity
import numpy as np
from scipy.spatial.distance import pdist


class Transcriptome_embedding:
    def __init__(self, embedding_index, embedding_matrix):
        self.embeddings_index = embedding_index
        self.embedding_matrix = embedding_matrix

    def search_relevant_embeddings_by_sample(self, samples):
        """
        This function takes a list of samples and returns the embeddings for those samples"""
        searching = self.embedding_matrix[self.embeddings_index.index.isin(samples),:]
        return searching

    def search_relevant_embeddings_by_series(self, series):
        """This function takes a list of series and returns the embeddings for those series
        """
        samples = self.embeddings_index[self.embeddings_index["series_id"].isin(series)].index
        searching = self.embedding_matrix[self.embeddings_index.index.isin(samples),:]
        return searching
    
    def obtain_samples_from_series(self, series):
        samples = self.embeddings_index[self.embeddings_index["series_id"].isin(series)].index
        return samples
    
    def mean_interpoint_distance(self, data, p = 1):
    # Calculate pairwise distances between points using Minkowski distance
        distances = pdist(data, metric='minkowski', p=p)

        # Calculate the mean distance
        mean_distance = np.mean(distances)
        scientific_notation = "{:.2e}".format(mean_distance)

        return scientific_notation
    
    def get_gse_from_transcriptome_index(self, ind):
        gse_id = self.embeddings_index.iloc[ind,:].to_list()[0]
        return gse_id
    
    def calculate_interpoint_distance_for_samples(self, samples, p):
        searching = self.search_relevant_embeddings_by_sample(samples)
        dist = self.mean_interpoint_distance(searching, p)
        return dist

    def calculate_interpoint_distance_for_series(self, series, p):
        searching = self.search_relevant_embeddings_by_series(series)
        dist = self.mean_interpoint_distance(searching, p)
        return dist
    
    def get_closest_transcriptional_studies(self, query_gse_id, k = 5):
        relevant_embedding_indices = np.where(self.embeddings_index["series_id"].isin([query_gse_id]))
        numpy_array = relevant_embedding_indices[0]
        flattened_array = numpy_array.flatten()
        query_embeddings = self.embedding_matrix[flattened_array,]
        
        cosine_similarities = cosine_similarity(self.embedding_matrix, query_embeddings)
        average_cosine_similarities = np.mean(cosine_similarities, axis=1)
        
        sorted_indices = np.argsort(average_cosine_similarities)[::-1]
        
        top_k = k
        results_data = []
    
        for i in range(top_k):
            index = sorted_indices[i]
            similarity = average_cosine_similarities[index]
            results_data.append({'Index': index, 'similarity_score': similarity})
    
        # Create a Pandas DataFrame from the results_data
        results_df = pd.DataFrame(results_data)
        results_df["gse_id"] = [self.get_gse_from_transcriptome_index(i) for i in results_df.Index]
        return results_df

    def get_closest_transcriptional_studies_by_sample(self, query_sample, k = 5):
        """this takes a (single) sample in a list and returns the k closest transcriptional studies"""
        embedding = self.search_relevant_embeddings_by_sample(query_sample)
        query_vector = embedding.reshape(1, -1)
        cosine_similarities = cosine_similarity(query_vector, self.embedding_matrix)
        closest_indices = np.argsort(cosine_similarities[0])[-k:][::-1]
        closest_samples = self.embeddings_index.iloc[closest_indices,:]
        return closest_samples



import pandas as pd
import anndata as ad
import decoupler as dc
import numpy as np
import archs4py as a4
import statsmodels.stats.multitest as smt
from tqdm import tqdm

class RNASeqAnalysis:
    def __init__(self, file):
        """
        Initializes a new instance of the `rnaseq_analysis` class.

        Args:
            file (str): The file path to be assigned to the `file` attribute.

        Returns:
            None
        """
        self.file = file

    def create_anndata_from_series(self, series):
        """Create an AnnData object with counts and metadata for a given series.

        Args:
            series (str): The name of the series.

        Returns:
            AnnData: An AnnData object containing the counts and metadata for the series.
        """
        series_counts = a4.data.series(self.file, series)
        metadata = a4.meta.series(self.file, series)
        test_ad = ad.AnnData(series_counts.T)
        test_ad.obs = metadata
        test_ad.var_names_make_unique()
        return test_ad

    def create_anndata_from_samples(self, samples):
        """Create an AnnData object with counts and metadata for the given samples.

        Args:
            samples (list): A list of sample names.

        Returns:
            AnnData: An AnnData object containing the counts and metadata for the samples.
        """
        sample_counts = a4.data.samples(self.file, samples)
        metadata = a4.meta.samples(self.file, samples)
        test_ad = ad.AnnData(sample_counts.T)
        test_ad.obs = metadata
        test_ad.var_names_make_unique()
        return test_ad
    
    def list_to_dc_geneset(self, list_test_geneset):
        """
        Convert a list of genes into a pandas DataFrame representing a geneset.

        Args:
            list_test_geneset (list): A list of genes.

        Returns:
            pandas.DataFrame: A DataFrame representing the geneset, with columns "genesymbol" and "geneset".

        """
        df = pd.DataFrame(list_test_geneset)
        df.columns = ["genesymbol"]
        df["geneset"] = "enrichment_score"
        return df

    def perform_enrichment_on_samples_batched(self, element_list, geneset, batch_size=2500):
        """
        Perform enrichment analysis on a list of elements in batches.

        Args:
            element_list (list): A list of elements to perform enrichment analysis on.
            geneset (str): The geneset to use for enrichment analysis.
            batch_size (int, optional): The size of each batch. Defaults to 5000.

        Returns:
            pandas.DataFrame: A combined DataFrame containing the results of enrichment analysis for all batches.
        """

        # Calculate the number of batches
        num_batches = (len(element_list) + batch_size - 1) // batch_size

        res_list = []

        # Iterate over batches
        for i in tqdm(range(num_batches)):
            start_index = i * batch_size
            end_index = (i + 1) * batch_size
            batch = element_list[start_index:end_index]

            # Apply your function to the current batch
            result = self.perform_enrichment_on_samples(batch, geneset)
            res_list.append(result)

        comb_df = pd.concat(res_list)
        return comb_df

    def perform_enrichment_on_series(self, query_series, gene_set):
        """
        This function accepts a list of series and a gene set and returns a DataFrame with the enrichment results.

        Parameters:
            query_series (list): A list of series.
            gene_set (str): The gene set to perform enrichment on.

        Returns:
            pandas.DataFrame: A DataFrame with the enrichment results.
        """
        
        samples_metadata = self.create_samples_from_series(query_series)
        df_res = self.perform_enrichment_on_samples_batched(samples_metadata.index, gene_set)
        return df_res
    
    def create_samples_from_series(self, series):
        """
        Create samples from a given series.

        Parameters:
        - series: The series to create samples from.

        Returns:
        - samples: The samples created from the series.
        """
        samples = a4.meta.series(self.file, series)
        return samples

    def perform_enrichment_on_samples(self, query_sample_list, gene_set): 
        """
        Perform enrichment analysis on a list of samples using a given gene set.

        Args:
            query_sample_list (list): A list of samples to perform enrichment analysis on.
            gene_set (list): A list of genes representing the gene set to use for enrichment analysis.

        Returns:
            pandas.DataFrame: A DataFrame containing the enrichment results.

        """
        
        current_ad = self.create_anndata_from_samples(query_sample_list)
        
        gene_set = self.list_to_dc_geneset(gene_set)
        
        dc.run_ora(
            mat=current_ad,
            net=gene_set,
            source='geneset',
            target='genesymbol',
            verbose=False, use_raw=False, min_n = 1,
        )

        results = current_ad.obsm['ora_estimate']
        results["pvals"] = current_ad.obsm['ora_pvals']
        results["p_corrected"] = smt.multipletests(results["pvals"], method='fdr_bh')[1]

        df_res = pd.concat([results, current_ad.obs], axis=1)

        return df_res





import pandas as pd
import numpy as np
from .rag_embedding import Rag_embedding
import requests

class Sample_to_series_map:
    def __init__(self, h5_path, rag_index_path, rag_embedding_path):
        self.geo_metadata = self.process_metadata(h5_path)
        self.rag_embedding_object = Rag_embedding(rag_index_path, rag_embedding_path) 
        
    def process_metadata(self, h5_path):
        x = pd.read_pickle(h5_path)
        x["samples"] = x.index
        x = x[['series_id', 'samples']].drop_duplicates()
        return x
    
    def return_series_for_samples(self, samps):
        df = self.geo_metadata[self.geo_metadata["samples"].isin(samps)]
        return df
    
    def retrieve_text(self, gse_id_list):
        text_list = []
        for i in gse_id_list:
            try:
                txt = self.rag_embedding_object.get_text_linked_to_gse(i)
                text_list.append(txt)
            except:
                pass
        return text_list
    
    def calculate_similarites_from_samples(self, query, samples):
        samples = self.return_series_for_samples(samples)
        experimental_text = self.retrieve_text(samples["series_id"])
        avg = self.rag_embedding_object.get_averages_between_queries(query, experimental_text)
        return avg

    def calculate_similarites_from_series(self, query, series_list):
        experimental_text = self.retrieve_text(series_list)
        avg = self.rag_embedding_object.get_averages_between_queries(query, experimental_text)
        return avg


class ARCHS4_API_query:
    def __init__(self, h5_path):
        self.geo_metadata = self.process_metadata(h5_path)
        
    def process_metadata(self, h5_path):
        x = pd.read_pickle(h5_path)
        x["samples"] = x.index
        x = x[['series_id', 'samples']].drop_duplicates()
        return x
        
    def create_data_dict(self, direction, geneset):
        if direction == "up":
            data = {
                "type" : "geneset",
                "direction" : "similar",
                "signatureName" : "example_query",
                "species" : "human",
                "upgenes" : geneset
            }
            return data
        else:
            data = {
                "type" : "geneset",
                "direction" : "similar",
                "signatureName" : "example_query",
                "species" : "human",
                "downgenes" : geneset
            }
            return data
            
    def extract_sig(self, direction, sig_list):
        URL = 'https://maayanlab.cloud/rookpy/signature'
        data = self.create_data_dict(direction, sig_list)
        r = requests.post(URL, json=data)
        samps = r.json()["samples"]
        to_search = ["GSM" + str(i) for i in samps]
        return to_search
    
    def return_relevant_series_and_samples(self, direction, sig_list):
        samps = self.extract_sig(direction, sig_list)
        df = self.geo_metadata[self.geo_metadata["samples"].isin(samps)]
        return df, samps


class MsigDB_store:
    def __init__(self, signature_df_path, signature_metadata_df_path):
        self.signature_metadata_df = pd.read_csv(signature_metadata_df_path)
        self.signature_df = self.load_gene_set_database(signature_df_path)
    
    def get_gene_set(self, genesetname):
        index_loc = np.where(self.signature_df["gene_set_name"].isin([genesetname]))[0][0]
        genelist = self.signature_df.iloc[index_loc,2].split("\t")
        return genelist

    def get_title_from_index(self, genesetname):
        index_loc = np.where(self.signature_metadata_df["standard_name"].isin([genesetname]))[0][0]
        title_long = self.signature_metadata_df.iloc[index_loc,5]
        title_short = self.signature_metadata_df.iloc[index_loc,6]
        return title_short, title_long
    
    def load_gene_set_database(self, path):
        x = pd.read_csv(path, header = None)
        x.columns = ["unparsed_geneset"]
        correctly_formatted = x["unparsed_geneset"].str.split("\t", n = 2, expand = True)
        correctly_formatted.columns = ["gene_set_name", "link", "geneset"]
        return correctly_formatted
    
    def list_genesets(self):
        gene_set_names = self.signature_metadata_df["standard_name"]
        return gene_set_names
    
    def get_gene_set_by_name(self, geneset_name):
        gene_list = self.get_gene_set(geneset_name)
        short_title, long_title = self.get_title_from_index(geneset_name)
        response = {"geneset" : gene_list, "short_title" : short_title, "long_title" : long_title}
        return response



import pandas as pd
from sklearn.metrics.pairwise import cosine_similarity
from sentence_transformers import SentenceTransformer
import numpy as np
import logging 
import sys

class Rag_embedding:
    def __init__(self, rag_index, rag_embedding_matrix):
        self.logger = logging.getLogger(__name__)
        H = logging.StreamHandler(sys.stdout)
        H.setLevel(logging.INFO)
        H.setFormatter(
            logging.Formatter(
                fmt="[%(asctime)s] %(levelname)s: %(message)s",
                datefmt="%d/%m/%Y ( %H:%M:%S )"
            ))
        self.logger.addHandler(H)


        self.model = SentenceTransformer("paraphrase-MiniLM-L6-v2")
        self.logger.info("Loaded SentenceTransformer model.")
        self.rag_embedding_index = rag_index
        self.rag_embedding_matrix = rag_embedding_matrix

    def get_gse_from_rag_index(self, ind):
        """Get the GSE ID from the RAG embedding index.

        Args:
            ind (int): The index of the GSE ID in the RAG embedding index.

        Returns:
            str: The GSE ID corresponding to the given index.
        """
        gse_id = self.rag_embedding_index.iloc[ind, :].to_list()[0]
        return gse_id

    def get_gse_text_from_rag_index(self, ind):
        """
        Retrieves the GSE text from the RAG index.

        Parameters:
        - ind (int): The index of the GSE text to retrieve.

        Returns:
        - gse_id_text (str): The GSE text corresponding to the given index.
        """
        gse_id_text = self.rag_embedding_index.iloc[ind, :].to_list()[1]
        return gse_id_text

    def get_text_linked_to_gse(self, gse_id_query):
        """
        Retrieves the text linked to a given GSE ID.

        Parameters:
        - gse_id_query (str): The GSE ID to query.

        Returns:
        - str: The text linked to the specified GSE ID.
        """
        gse_id_text = self.rag_embedding_index[
            self.rag_embedding_index["series_id"].isin([gse_id_query])
        ]["rag_text"].to_list()[0]
        return gse_id_text

    def get_closest_semantic_studies(self, query_gse_id, k=5):
        """
        Retrieves the closest semantic studies based on a given query GSE ID.

        Parameters:
            query_gse_id (str): The GSE ID to query against for expansion.
            k (int): The number of closest studies to retrieve. Default is 5.

        Returns:
            pandas.DataFrame: A DataFrame containing the closest studies along with their similarity scores.

        """
        relevant_embedding_indices = np.where(
            self.rag_embedding_index["series_id"].isin([query_gse_id])
        )
        numpy_array = relevant_embedding_indices[0]
        flattened_array = numpy_array.flatten()
        query_embeddings = self.rag_embedding_matrix[flattened_array,]

        cosine_similarities = cosine_similarity(
            self.rag_embedding_matrix, query_embeddings
        )
        average_cosine_similarities = np.mean(cosine_similarities, axis=1)

        sorted_indices = np.argsort(average_cosine_similarities)[::-1]

        top_k = k
        results_data = []

        for i in range(top_k):
            index = sorted_indices[i]
            similarity = average_cosine_similarities[index]
            results_data.append({"Index": index, "similarity_score": similarity})

        # Create a Pandas DataFrame from the results_data
        results_df = pd.DataFrame(results_data)
        results_df["gse_id"] = [
            self.get_gse_from_rag_index(i) for i in results_df.Index
        ]
        results_df["gse_id_text"] = [
            self.get_gse_text_from_rag_index(i) for i in results_df.Index
        ]
        # results_df["similarity_source"] = "transcriptome"
        return results_df
    
    def query_rag(self, query, k=5):
        """
        Query the RAG (Retrieval-Augmented Generation) model with a given query.

        Args:
            query (str): The query text.
            k (int, optional): The number of top results to retrieve. Defaults to 5.

        Returns:
            pandas.DataFrame: A DataFrame containing the top k results with columns 'similarity_score', 'gse_id', 'gse_id_text', and 'similarity_source'.
        """
        query_embedding = self.model.encode(query).reshape(1, -1)
        cosine_similarities = cosine_similarity(
            self.rag_embedding_matrix, query_embedding
        )
        cosine_distances = 1 - cosine_similarities
        sorted_indices = np.argsort(cosine_distances.flatten())

        if len(sorted_indices) < k:
            k = len(sorted_indices)

        top_k = k
        results_data = []

        for i in range(top_k):
            index = sorted_indices[i]
            distance = cosine_distances[index]
            results_data.append({"Index": index, "similarity_score": distance[0]})

        # Create a Pandas DataFrame from the results_data
        results_df = pd.DataFrame(results_data)
        results_df["gse_id"] = [self.get_gse_from_rag_index(i) for i in results_df.Index]
        results_df["gse_id_text"] = [
            self.get_gse_text_from_rag_index(i) for i in results_df.Index
        ]
        results_df["similarity_source"] = "semantic"

        return results_df.drop("Index", axis=1)
    
    def get_averages_between_queries(self, query, list_of_queries):
        """
        Calculates the average cosine similarity between a query and a list of queries.

        Parameters:
        query (str): The query string.
        list_of_queries (list): A list of query strings.

        Returns:
        float: The average cosine similarity between the query and the list of queries.
        """
        query_embedding = self.model.encode([query])[0]
        list_embeddings = self.model.encode(list_of_queries)
        
        query_embedding = query_embedding.reshape(1, -1)
        list_embeddings = list_embeddings.reshape(len(list_of_queries), -1)

        # Calculate cosine similarity using scikit-learn
        cosine_scores = cosine_similarity(query_embedding, list_embeddings)

        # Calculate average similarity
        average_similarity = np.mean(cosine_scores)
        return average_similarity



import pandas as pd
import decoupler as dc
import numpy as np
import statsmodels.stats.multitest as smm

class Transcriptome_enrichment:
    def __init__(self, h5ad_obj):
        """
        Initializes the Enrichment class, which stores representative transcriptomes

        Parameters:
        - h5ad_obj: An h5ad object representing the data to be enriched.
        """
        self.memmap_adata = h5ad_obj
   
    def list_to_dc_geneset(self, list_test_geneset):
        """
        Convert a list of genes to a pandas DataFrame representing a gene set.

        Args:
            list_test_geneset (list): A list of gene symbols.

        Returns:
            pandas.DataFrame: A DataFrame with two columns: 'genesymbol' and 'geneset'.
                              The 'genesymbol' column contains the gene symbols from the input list,
                              and the 'geneset' column is set to 'query_gene_set' for all rows.

        """
        df = pd.DataFrame(list_test_geneset)
        df.columns = ["genesymbol"]
        df["geneset"] = "query_gene_set"
        return df
    
    def list_to_dc_geneset_dictionary(self, list_test_geneset_dict):
        """
        Convert a list of test geneset dictionaries into a pandas DataFrame.

        Args:
            list_test_geneset_dict (list): A list of test geneset dictionaries.

        Returns:
            pandas.DataFrame: A DataFrame containing the genesets and their corresponding genes.

        """
        geneset_df_list = []
        for i, j in list_test_geneset_dict.items():
            df = pd.DataFrame(j)
            df.columns = ["genesymbol"]
            df["geneset"] = i
            geneset_df_list.append(df)
        return pd.concat(geneset_df_list)
    
    def run_decouplr_on_memmaped_adata_with_samples(self, genelist, samples_list):
        """
        Runs the decouplr algorithm on a memmaped AnnData object with a specified list of samples.

        Args:
            genelist (dict or list): The gene list to use for the decouplr algorithm. It can be either a dictionary
                representing a geneset or a list of genes.
            samples_list (list): The list of sample names to include in the analysis.

        Returns:
            pandas.DataFrame: A DataFrame containing the results of the decouplr algorithm, including the p-values,
            estimates, and corrected p-values (if applicable).
        """
        if isinstance(genelist, dict):
            geneset = self.list_to_dc_geneset_dictionary(genelist)
        elif isinstance(genelist, list):
            geneset = self.list_to_dc_geneset(genelist)
        else:
            print(genelist)

        adata = self.memmap_adata
        query_adata = adata[adata.obs.index.isin(samples_list)].to_memory()
        query_adata.var_names_make_unique()
        np.nan_to_num(query_adata.X, copy=False, nan=0.0, posinf=0.0, neginf=0.0)

        dc.run_ora(
            mat=query_adata,
            net=geneset,
            source='geneset',
            target='genesymbol',
            verbose=False,
            use_raw=False,
            min_n=1
        )

        if query_adata.obsm["ora_pvals"].shape[1] > 1:
            return query_adata.obsm["ora_pvals"], query_adata.obsm["ora_estimate"]
        else:
            res = pd.concat([query_adata.obsm["ora_pvals"], query_adata.obsm["ora_estimate"]], axis=1)
            res.columns = ["pvals", "estimate"]
            _, p_corrected, _, _ = smm.multipletests(res['pvals'], method='fdr_bh')
            res['p_corrected'] = p_corrected
            return res
    

